/* eslint-disable @typescript-eslint/prefer-ts-expect-error, @typescript-eslint/ban-ts-comment */
import listen from 'test-listen';
import fetch from 'isomorphic-unfetch';
import _$http from 'http';
const {
  createServer
} = _$http;
import _$url from 'url';
const {
  parse: parseUrl
} = _$url;
import _$cookie from 'cookie';
const {
  parse: parseCookieHeader
} = _$cookie;
let apiResolver = null;

// ? The result of the calling function is memoized, so this function will only
// ? be invoked the first time this script is imported.

/* istanbul ignore next */
const tryImport = path => e => {
  var _tryImport$importErro;

  (tryImport.importErrors = (_tryImport$importErro = tryImport.importErrors) !== null && _tryImport$importErro !== void 0 ? _tryImport$importErro : []).push(e); // ? Node12 does not support dynamic imports, so fall back to require for now

  return (typeof __webpack_require__ === 'function' ? __non_webpack_require__ : require)(path);
};
/**
 * The parameters expected by `testApiHandler`.
 */


/**
 * Uses Next's internal `apiResolver` to execute api route handlers in a
 * Next-like testing environment.
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export async function testApiHandler({
  requestPatcher,
  responsePatcher,
  paramsPatcher,
  params,
  url,
  handler,
  test
}) {
  let server = null;

  try {
    /* istanbul ignore next */
    if (!apiResolver) {
      // ? The following is for next@>=11.1.0:
      // @ts-ignore: conditional import for earlier next versions
      ({
        apiResolver
      } = await import('next/dist/server/api-utils.js') // ? The following is for next@<11.1.0 >=9.0.6:
      .catch(tryImport('next/dist/next-server/server/api-utils.js')) // ? The following is for next@<9.0.6 >= 9.0.0:
      .catch(tryImport('next-server/dist/server/api-utils.js')).catch(e => (tryImport.importErrors.push(e), {
        apiResolver: null
      })));

      if (!apiResolver) {
        const importErrors = tryImport.importErrors.map(e => e.message.split(/(?<=')( imported)? from '/)[0].split(`\nRequire`)[0]).join('\n    - ');
        tryImport.importErrors = []; // prettier-ignore

        throw new Error(`failed to import api-utils.js` + `\n\n  This is usually caused by:` + `\n\n    1. Using npm@<7 and/or node@<15, which doesn't install peer deps automatically (review install instructions)` + `\n    2. NTARH and the version of Next.js you installed are actually incompatible (please submit a bug report)` + `\n\n  Failed import attempts:` + `\n\n    - ${importErrors}`);
      }
    }

    const localUrl = await listen(server = createServer((req, res) => {
      try {
        /* istanbul ignore next */
        if (typeof apiResolver != 'function') {
          throw new Error('assertion failed unexpectedly: apiResolver was not a function');
        }

        url && (req.url = url);
        requestPatcher && requestPatcher(req);
        responsePatcher && responsePatcher(res);
        const finalParams = { ...parseUrl(req.url || '', true).query,
          ...params
        };
        paramsPatcher && paramsPatcher(finalParams);
        /**
         *? From Next.js internals:
         ** apiResolver(
         **    req: IncomingMessage,
         **    res: ServerResponse,
         **    query: any,
         **    resolverModule: any,
         **    apiContext: __ApiPreviewProps,
         **    propagateError: boolean
         ** )
         */

        void apiResolver(req, res, finalParams, handler, // eslint-disable-next-line @typescript-eslint/no-explicit-any
        undefined, false);
      } catch (e) {
        /* istanbul ignore next */
        // ? Prevent tests that crash the server from hanging
        !res.writableEnded && res.end();
        /* istanbul ignore next */
        // ? Allow the exception to bubble naturally

        throw e;
      }
    }));
    await test({
      fetch: async init => {
        return fetch(localUrl, init).then(res => {
          res.cookies = [res.headers.raw()['set-cookie'] || []].flat().map(header => Object.entries(parseCookieHeader(header)).reduce((obj, [k, v]) => Object.assign(obj, {
            [String(k)]: v,
            [String(k).toLowerCase()]: v
          }), {}));
          return res;
        });
      }
    });
  } finally {
    var _server;

    (_server = server) === null || _server === void 0 ? void 0 : _server.close();
  }
}